--[[
///////////////////////////////////////////////////////////////////////////////
 'TwinTail LDE'

 Copyright(c) 2005-2006, ema.
///////////////////////////////////////////////////////////////////////////////
]]--

-------------------------------------------------------------------------------
require("pad.lc")
require("editctrl.lc")
require("jinput.lc")
require("filer.lc")
require("osk/twintail.lc")
--require("osk/game.lc")


-------------------------------------------------------------------------------
Editor = {
  BaseDir   = System.currentDirectory()
, KeyType   = 0
, Frame     = 0
, Skin      = Image.load("skin.png")
, OnWindows = string.find(System.currentDirectory(), "/cygdrive/")
, MenuList = {
    "New"
  , "Open..."
  , "Save"
  , "Save as..."
  , "Close"
  , "USB connect"
  , "Exec as LUA"
  , "Exit"
  }
}

--dofile "ttlde.cfg"
dofile "ttlde.dic"


-------------------------------------------------------------------------------
function Editor.new(scr)
  if scr == nil then scr = screen end
  local c = setmetatable({
    Available  = true
  , ClipBoard  = ""
  , Target     = scr
  , iBuf       = 1
  , Buffers    = {}
  , FileName   = {}
  , GlobalState = {}
  , LoadedState = {}
  , timer = Timer.new()
  , OldTick = 0
  }, { __index = Editor })

  for Key, Value in pairs(_G) do
    c.GlobalState[Key] = Value
  end
  for Key, Value in pairs(_LOADED) do
    c.LoadedState[Key] = Value
  end

  c:New()
  c.timer:start()
  return c
end


-------------------------------------------------------------------------------
function Editor:RestoreState()
  _LOADED = {}
  for Key, Value in pairs(_G) do
    if Value ~= _G then
      _G[Key] = self.GlobalState[Key]
    end
  end
  for Key, Value in pairs(self.LoadedState) do
    _LOADED[Key] = Value
  end
  collectgarbage()
end

-------------------------------------------------------------------------------
function Editor:EditOneLine(msg, width, src)
  local oskActive = false
  local osk = Osk.new()
  local SubBuffer = EditControl.new(self)
  local jmode = false

  SubBuffer.MultiLine = false

  if src then
    SubBuffer.Buffer[1] = src
    SubBuffer.CurPos = string.len(src)+1
  end

  while true do
    Pad:Update()
    self:Draw()

    if not oskActive then
      if Pad.tR1 then
        oskActive = true
        osk = Osk.new()
      end
    end

    if oskActive then
      local In, cont = osk:Update()
      osk:Draw(SubBuffer.csrx, SubBuffer.csry+10)
      if jmode then In = Jinput:AddChar(In) end
      SubBuffer:Keyin(In)
      if Pad.tSt and not Jinput:IsActive() then jmode = not jmode end
      if not cont then oskActive = false end
    elseif Jinput:IsActive() then
      SubBuffer:Keyin(Jinput:Proc())
    else
      SubBuffer:EditProc()
      if not Pad.L1 then
        if Pad.tRr then return SubBuffer.Buffer[1] end
        if Pad.tRd then return end
      end
    end
    SubBuffer:DrawEditBox(width, msg)

    if Jinput:IsActive() then
      Jinput:Draw(SubBuffer.csrx, SubBuffer.csry)
    end

    self:Flip()
  end
end


-------------------------------------------------------------------------------
function Editor:Confirm()
  if self.Buffers[self.iBuf].Modified then
    return self:MsgBox("Discard?", 1) == 0
  else
    return true
  end
end

-------------------------------------------------------------------------------
function Editor:DrawBg()
  local this = self.Buffers[self.iBuf]
  local Target = self.Target
  Target:fillRect(30, 16, 442, 256, this.Colors.Bg)
  Target:blit(0, 0, self.Skin, 0, 0, 480, 16, false)
  Target:blit(0, 16, self.Skin, 0, 16, 30, 256, false)
  Target:blit(472, 16, self.Skin, 472, 16, 8, 256, false)

  if not self.OnWindows then
    local t = System.powerGetBatteryLifePercent() / 4
    if t<1 then t = self.Frame&16*3/2 end
    Target:blit(448, 4, self.Skin, 440, 24, 24, 8)
    Target:blit(448, 4, self.Skin, 440, 16, t, 8)
    t = System.getFreeMemory() / 1048576
    if t<2 then t = self.Frame&16*3/2 end
    Target:blit(420, 4, self.Skin, 440, 24, 24, 8)
    Target:blit(420, 4, self.Skin, 440, 16, t, 8)
  end
  Jdraw:print("["..self.iBuf.."] "..
    string.sub((this.FileName or "[untitled]"), -70),
    8, 3, Target)
end

-------------------------------------------------------------------------------
function Editor:Draw()
  self:DrawBg()
  self.Buffers[self.iBuf]:Draw()
end

-------------------------------------------------------------------------------
function Editor:MsgBox(msg, lv)
  local this = self.Buffers[self.iBuf]

  while true do
    Pad:Update()
    self:Draw()
    self:DrawBox(160, 100, 160, 100, "Confirm")
    self.Target:print(170, 120, msg)
    self.Target:print(170, 150, "Circle = Yes/OK")
    if lv>0 then self.Target:print(170, 160, "Cross  = No") end
    if lv>1 then self.Target:print(170, 170, "Square = Cancel") end
    self:Flip()
    if Pad.tRr then return 0 end
    if lv>0 and Pad.tRd then return 1 end
    if lv>1 and Pad.tRl then return 2 end
  end
end


-------------------------------------------------------------------------------
function Editor:DrawBox(x, y, w, h, s)
  local Target = self.Target
  local Skin = self.Skin
  Target:fillRect(x, y, w, h)
  Target:blit(x, y, Skin, 0, 0, w-8, 16)
  Target:blit(x+w-8, y, Skin, 472, 0, 8, 16)
  Target:blit(x+1, y+16, Skin, 0, 16, w-2, h-17)
  if s then Target:print(x+4, y+4, s) end
end


-------------------------------------------------------------------------------
function Editor:MenuMain(List, Title)
  local Scroll = 1
  local Sel=1
  local n=table.getn(List)
  local this = self.Buffers[self.iBuf]
  local unselcol = Color.new(160,160,160)
  local selcol = Color.new(0,0,0)

  while true do
    Pad:Update()
    self:Draw()

    if Pad.tRd then return nil end
    if Pad.tRr then return Sel end

    Sel = Sel + Pad:GetVrep()
    if Sel < 1 then Sel = n end
    if Sel > n then Sel = 1 end
    if Scroll > Sel     then Scroll = Sel     end
    if Scroll < Sel - 8 then Scroll = Sel - 8 end

    self:DrawBox(120, 100, 240, 100, Title)

    for i=1,9 do
      local iItem = i+Scroll-1
      local item = List[iItem]
      if item then
        local col
        if iItem == Sel then col = selcol else col = unselcol end
        self.Target:print(130, 113+i*8, item, col)
      end
    end
    self:Flip()
  end
end


-------------------------------------------------------------------------------
function Editor:ExecLua()
  local result,err = loadstring(table.concat(self.Buffers[self.iBuf].Buffer, "\n"))
  if not result then
    self:EditOneLine("Compile error", 80, err)
  else
    local OrgDir = System.currentDirectory()
    _LOADED = {}
    result, err = pcall(result)
    System.currentDirectory(OrgDir)
    self:RestoreState()
    if not result then
      self:EditOneLine("Runtime error", 80, err)
    else
      self:MsgBox("done.", 0)
    end
  end
end


-------------------------------------------------------------------------------
function Editor:ViewPicture(Name)
  local bg  = Image.createEmpty(480, 272)
  local pic = Image.load(Name)
  local w = 240 - pic:width()/2
  local h = 136 - pic:height()/2
  local x, y = 0, 0

  local c1 = Color.new(210, 210, 210)
  local c2 = Color.new(240, 240, 240)

  bg:fillRect(0, 0, 480, 272, c2)
  for i=0,464,16 do
    for j=0,256,16 do
      bg:fillRect(i, j, 8, 8, c1)
      bg:fillRect(i+8, j+8, 8, 8, c1)
    end
  end

  repeat
    Pad:Update()
    x = x + Pad:GetHrep()
    y = y + Pad:GetVrep()
    self.Target:blit(0, 0, bg, 0, 0, 480, 272, false)
    self.Target:blit(w+x, h+y, pic)
    self:Flip()
  until Pad.tRd
end


-------------------------------------------------------------------------------
function Editor:New()
  local t = EditControl.new(self)
  table.insert(self.Buffers, t)
  self.iBuf = table.getn(self.Buffers)
  return t
end


-------------------------------------------------------------------------------
function Editor:Open()
  Filer:Init(self.Skin, self.Buffers[self.iBuf].Colors)
  local NewName = Filer:Exec()
  if NewName then
    if string.find(NewName, "[pP][nN][gG]$") or
       string.find(NewName, "[jJ][pP][gG]$") then
      self:ViewPicture(NewName)
    else
      self:New()
      self.Buffers[self.iBuf]:Load(NewName)
      self.Buffers[self.iBuf].FileName = NewName
    end
  end
end

-------------------------------------------------------------------------------
function Editor:Close()
  table.remove(self.Buffers, self.iBuf)
  self.iBuf = table.getn(self.Buffers)
  if self.iBuf == 0 then self:New() end
end


-------------------------------------------------------------------------------
function Editor:FileMenu()
  local Sel = self:MenuMain(self.MenuList, "File")
  local this = self.Buffers[self.iBuf]

  if Sel==1 then self:New() end
  if Sel==2 then self:Open() end

  if Sel==3 then
    if not this.FileName then
      Sel=4
    else
      this:Save(this.FileName)
    end
  end

  if Sel==4 then
    local NewName =
      self:EditOneLine("Input file name", 30, this.FileName)
    if NewName then
      this:Save(NewName)
      this.FileName = NewName
    end
  end

  if Sel==5 then
    if self:Confirm() then self:Close() end
  end

  if Sel==6 then
    System.usbDiskModeActivate()
    self:MsgBox("USB connecting", 0)
    System.usbDiskModeDeactivate()
  end

  if Sel==7 then
    if self:MsgBox("Sure?", 1) == 0 then
      self:ExecLua()
    end
  end

  if Sel==8 then
    if self:Confirm() then
      if self:MsgBox("Exit?", 1) == 0 then
        self.Available = false
      end
    end
  end
end


-------------------------------------------------------------------------------
function Editor:PluginMenu()
  local this = self.Buffers[self.iBuf]
  local ef
  local Dir = "/plugin"

  while true do
    local FileList = {}
    local PluginNameList = {}
    if self.OnWindows then
      FileList = {
        {}, {}
      , {name="1000_Search....lua"}
      , {name="2000_Jump to....lua"}
      , {name="3000_Color Scheme.lua"}
      }
    else
      FileList = System.listDirectory(self.BaseDir .. Dir)
    end

    -- remove "." and ".."
    table.remove(FileList, 1)
    table.remove(FileList, 1)
    table.sort(FileList,
      function(a, b)
        if a.directory == b.directory then return a.name < b.name end
        return b.directory
      end
    )

    for idx,file in FileList do
      if file.directory then
        table.insert(PluginNameList, "/" .. file.name)
      else
        local ul  = string.find(file.name, "_") or 0
        local dot = string.find(file.name, "lua$") or 0
        table.insert(PluginNameList, string.sub(file.name, ul+1, dot-2))
      end
    end

    local Sel = self:MenuMain(PluginNameList, "Plugin")
    if not Sel then return end

    if FileList[Sel].directory then
      Dir = Dir .. PluginNameList[Sel]
    else
      local fln = self.BaseDir .. Dir .. "/" .. FileList[Sel].name
      local func, err = loadfile(fln)
      if not func then
        self:EditOneLine("Compile error", 40, err)
      else
        dofile(fln)
        PluginFunc(self, this)
      end
      return
    end
  end
end


-------------------------------------------------------------------------------
function Editor:Flip()
--  screen.waitVblankStart()
  self.Frame = self.Frame + 1
  screen.flip()
end


-------------------------------------------------------------------------------
function Editor:SelectGlobal()
  local White = Color.new(255, 255, 255)
  local Red   = Color.new(255,   0,   0)
  local Yellow= Color.new(255, 255,   0)
  local Cyan  = Color.new(  0, 255, 255)
  local Path  = {}
  local ret
  local FullPath
  local pfxmode = true
  local Prefix = { [false]=".", [true]=":" }

  local SortProc = function(a, b)
    if a.Type == b.Type then
      return a.Name < b.Name
    end
    if a.Type == "table" then return true  end
    if b.Type == "table" then return false end
    if a.Type == "function" then return true  end
    if b.Type == "function" then return false end
    return a.Type < b.Type
  end

  while not ret do
    local Sel = 1
    local tbl = _G
    local ValList = { }
    local change = false
    local nFile = 0
    local Scroll = 0

    FullPath = ""
    for Lv, Name in pairs(Path) do
      if Name == "[true]" then
        tbl = tbl[true]
      elseif Name == "[false]" then
        tbl = tbl[false]
      elseif string.find(Name, "[", 1, true) then
        tbl = tbl[tonumber(string.sub(Name, 2, -2))]
      else
        tbl = tbl[Name]
        if FullPath ~= "" then Name = "." .. Name end
      end
      FullPath = FullPath .. Name
    end

    if tbl == nil then ret = "" ; break end

    for Key, Value in pairs(tbl) do
      local elem = { Key=Key, Name=Key, Type=type(Value) }
      if type(Key) == "string" then
        if string.find(Key, "_") ~= 1 then
          table.insert(ValList, elem)
          nFile = nFile + 1
        end
      elseif type(Key) == "number" then
        elem.Name = "[" .. Key .. "]"
        table.insert(ValList, elem)
        nFile = nFile + 1
      elseif type(Key) == "boolean" then
        if Key then
          elem.Name = "[true]"
        else
          elem.Name = "[false]"
        end
        table.insert(ValList, elem)
        nFile = nFile + 1
      end
    end

    table.sort(ValList, SortProc)

    while not change do
      Pad:Update()
      self.Target:clear()
      self.Target:print(0, 0, "Variable path   = ["..FullPath.."]", White)
      self.Target:print(0, 8, "Function prefix = ["..Prefix[pfxmode].."]", White)

      Sel = Sel + Pad:GetVrep()
      if Sel < 1 then Sel = nFile end
      if Sel > nFile then Sel = 1 end
      if Scroll > Sel - 10 then Scroll = Sel - 10 end
      if Scroll < Sel - 20 then Scroll = Sel - 20 end

      if Pad.tLl or Pad.tLr then pfxmode = not pfxmode end
      if Pad.tRu then ret = "" ; break end
      if Pad.tRd then
        if table.getn(Path) == 0 then
          ret=""
        else
          table.remove(Path)
        end
        change = true
      end

      for i=0, 30 do
        local iFile = i+Scroll
        local elem = ValList[iFile]
        local y = i*8 + 16

        if elem then
          local col
          local Type = elem.Type
          local Name = elem.Name
          if Type == "table" then
            if Pad.tRr and iFile == Sel then
              table.insert(Path, Name)
              change = true
            end
            col = Cyan
          elseif Type == "function" then
            if Pad.tRr and iFile == Sel then
              ret = Name .. "()"
              change = true
            end
            col = Yellow
          else
            if Pad.tRr and iFile == Sel then
              ret = Name
              change = true
            end
            col = White
          end
          if Pad.R1 and Pad.tRl and iFile == Sel then
            tbl[elem.Key] = nil
            change = true
          end

          if iFile == Sel then col = Red end
          self.Target:print(8, y, Name, col)
          self.Target:print(380, y, "[" .. Type .. "]", col)
        end
      end

      screen.waitVblankStart()
      screen.flip()
    end
  end
  if ret ~= "" then
    if not string.find(ret, "[", 1, true) then
      local pfx = "."
      if string.find(ret, "()", 1, true) then
        pfx = Prefix[pfxmode]
      end
      if FullPath ~= "" then ret = pfx .. ret end
    end
    local this = self.Buffers[self.iBuf]
    this:InsertString(FullPath .. ret)
    this.CopyInfo.eLine = this.CurLine
    this.CopyInfo.ePos  = this.CurPos
    this:UpdateSelection()
  end
end


-------------------------------------------------------------------------------
function Editor:QuickInput()
  local Category, Sel = 1, 1
  local this = self.Buffers[self.iBuf]
  local moved = false

  this.CopyInfo.sLine = this.CurLine
  this.CopyInfo.sPos  = this.CurPos
  this.CopyInfo.eLine = this.CurLine
  this.CopyInfo.ePos  = this.CurPos
  this:UpdateSelection()
  this:SetUndo()

  while true do
    Pad:Update()
    self:Draw()

    if not Pad.Ru then
      if not moved then self:SelectGlobal() end
      break
    end

    local h = Pad:GetHrep()
    local v = Pad:GetVrep()

    if h ~= 0 then
      local n = table.getn(self.Dictionary)
      Category = Category + h
      if Category < 1 then Category = n end
      if Category > n then Category = 1 end
      Sel, v = 0, 1
      moved = true
    end

    if v ~= 0 then
      local n = table.getn(self.Dictionary[Category])
      Sel = Sel + v
      if Sel < 1 then Sel = n end
      if Sel > n then Sel = 1 end

      this:DeleteSelection()
      this:InsertString(self.Dictionary[Category][Sel])

      this.CopyInfo.eLine = this.CurLine
      this.CopyInfo.ePos  = this.CurPos
      this:UpdateSelection()
      moved = true
    end

    self:Flip()
  end
end

-------------------------------------------------------------------------------
-- Main routine
do
  local editor = Editor.new()
  local bufchg = false
  local jmode = false
  local oskActive = false
  local osk

  Jinput:Init()

  editor.Buffers[editor.iBuf]:Load(editor.BaseDir .. "/readme.txt")

  while editor.Available do
    Pad:Update()
    editor:Draw()

    if not oskActive then
      if Pad.tR1 then
        oskActive = true
        osk = Osk.new()
      end
    end

    if oskActive then
      local ctrl = editor.Buffers[editor.iBuf]
      local In, cont = osk:Update()
      osk:Draw(ctrl.csrx, ctrl.csry+10)
      if jmode then In = Jinput:AddChar(In) end
      ctrl:Keyin(In)
      if Pad.tSt and not Jinput:IsActive() then jmode = not jmode end
      if not cont then oskActive = false end
    elseif Jinput:IsActive() then
      editor.Buffers[editor.iBuf]:Keyin(Jinput:Proc())
    else
      if not Pad.L1 and Pad.tRu then
        editor:QuickInput()
      end
      if Pad.tSt then editor:FileMenu() end
      if Pad.Sl then
        local n = table.getn(editor.Buffers)
        local h = Pad:GetHrep()
        if h ~= 0 then
          bufchg = true
          editor.iBuf = editor.iBuf + h
          if editor.iBuf < 1 then editor.iBuf = n end
          if editor.iBuf > n then editor.iBuf = 1 end
        end
      elseif Pad.rSl then
        if not bufchg then editor:PluginMenu() end
        bufchg = false
      else
        editor.Buffers[editor.iBuf]:EditProc()
      end
    end

    if Jinput:IsActive() then
      local ctrl = editor.Buffers[editor.iBuf]
      Jinput:Draw(ctrl.csrx, ctrl.csry)
    end

    editor:Flip()
  end
end